//LinkList class by Jason Winnebeck  (helped by Cylord and many others!)
//This class is freeware
//VERSION 1.9 FINAL 5/25/99

//I've had a lot of trouble with #TDLinkList.  It really requires knowledge
// of the internal workings of LinkList.cc, so I strongly advice against it.

#ifndef __LINKLIST_H_     //only include this class once
#define __LINKLIST_H_

#ifndef NULL
#define NULL 0
#endif

#define linklistver 1.9  //LinkList version number
#define linkliststr "LinkList class by Jason Winnebeck"
                          //Author string

template <class linkdata> //template to fit any data type
struct Link {             //Link structure, 1 link per data
  linkdata *data;         //Data stored in the link
  Link *next;             //pointer to next link
};

template <class linkdata>
class LinkList {
private:
  Link<linkdata> *first; //pointer to the first element of the list
  Link<linkdata> *next;  //For the GetNext() routines
  Link<linkdata> *prev;  //For the _DeleteThis() routine
  int len;               //Length of the LinkList

  Link<linkdata> *GetIndx(int indx);
    //Many funcs need a pointer to a link, so this shortens redundancies

#ifndef TDLinkList
  //If you want LinkLists in LinkLists, #define TDLinkList, but keep in mind
  // not to copy linklists yourself.
  operator = (LinkList<linkdata>);
  LinkList<linkdata>(const LinkList<linkdata> &list);
    //Bodies are never defined for these because copying linklists would
    // require a member-by-member copy, which are usually not required in a
    // program.  Copy manually if you HAVE to get a copy
#endif

public:
  LinkList();            //constructor
  ~LinkList();           //destructor

//*** Set array data functions ***
  void Clear();
    //deletes everything currently in the list
  void AddItem(const linkdata &data);
  void AddItem(linkdata *data);
    //Adds to start of the list (dual methods)
  bool DeleteItem(int indx);
    //Delete a link
  void Swap(LinkList<linkdata> &l);
    //Swaps two LinkLists, deleting or adding nothing.  Calling object becomes l, l becomes obj

//***   Array info functions   ***
  inline int Length() {return len;}
    //Returns the length of the array
  inline bool IsAtEnd() {return (next->next == NULL);}
    //Returns true if the list is on the LAST element
  inline bool IsOverEnd() {return (next == NULL);}
    //Returns true if the list is one PAST last element (over-run)

//*** Retrieve array data funcs***
  linkdata *GetItem(int indx);
    //Direct Item Manipulation (SLOW ACCESS!!!)
  bool GetItem(int indx, linkdata& dat);
    //Error checking and for data, not ptrs
  linkdata& operator [] (int indx);
    //"Traditional" array access, returns ref, can be used as lvalue
    //out of index returns last element in array!!! (can't return NULL by ref)
  void SetItem(int indx, const linkdata& dat);
    //Sets a link to dat
  linkdata *MakeLinear();
    //Returns a linear array of Length() of linkdata, copying all data
    //Delete memory with delete[] after use!
    //Returns NULL if the list is empty

//***  Sequential Access funcs ***
  inline void SetToStart() {next = first; prev = NULL;}
    //Sequential Item Access (FAST)
  bool GetNext(linkdata &dat);
    //Gets next element and increments position

//***   Adv/optimized functs   ***
  void _MoveData(linkdata *dat);
  //This is just like AddItem, but DOES NOT COPY DATA IN!  This is a VERY
  // dangerous function!  Keep in mind that any change to the ptr after a call
  // to _MoveData will change data as it lays in the array (could also be
  // beneficial, as _GetNext() is... optimization purposes)
  //Also once the data is moved over the memory allocation is now LinkList's
  // responsibility... you have no need to delete the memory, it is even
  // recommended that you consider that the temporary pointer you used never
  // existed after you move it.

  void _AddHere(linkdata *dat);
  //This function is for creating sorted linklists. This function copies the
  // data into the linklist before the element.  A good method is to proceed
  // until _GetNext() returns the element greater, then call _AddHere().  You
  // do not need to increment the for variable, because the addition occurs
  // before next.

  void _MoveHere(linkdata *dat);
  //Same as Addhere, but moves instead of copies.

  void _DeleteThis();
  //***WARNING!!!*** For speed, assumes the array has AT LEAST 1 element!!!
  // (bypass this by using a loop, for(c=0; c<len; c++) {} ) which would not
  // enter the loop unless there is elements (and therefore calcs to do))
  //Deletes the link currently at, so you can do all list functions in one
  // sequential pass through the list to make it speed-efficent.  This will
  // delete the link and advance, using _GetNext will now return the next.
  // If this is the last element, the list will run off (and your loop
  // SHOULD terminate.)  This will be fine as long as you call SetToStart()
  // before your next access to start at the beginning.

  inline void _Advance() {
    prev = next;
    next = next->next;
  }
  //Must call this to get next element when
  // using the optimized _GetNext() ONLY!

  inline linkdata *_GetNext() {
    return next->data;
  }
  //Optimized GetNext(no range check) (Fastest)
  //****WARNING**** THIS VARIABLE IS NOT COPIED!!!
  // however, this is a good way to take the data and process it,
  // and change the data in the process (thereby changing the data in the arr)
};

//Rem this if using in a project makefile
#include "linklist.cc"

#endif //__LINKLIST_H_
